/*
 * Png stitcher.
 *
 * Png stitcher is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Png stitcher is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ascad; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
/* Copyright (C) 2007 Andrei Krger */

// #define deb(fmt, args...) printf("%s: " fmt, __FUNCTION__, ##args)
#define deb(fmt, args...) while(0) {}

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>
#include <png.h>
#include <setjmp.h>
#include "aspnger.h"

#define SKIP_TOP_ROWS 18
#define SKIP_BOTTOM_ROWS 32

int ispng(FILE *f)
{
  unsigned char       buf[8];

  if (!f)
    return 0;
  fread(buf, 1, 8, f);
  rewind(f);
  return (int)png_check_sig(buf, 8);
}

void png_set_filters(png_struct *png)
{
  png_set_strip_16(png);
  png_set_gray_to_rgb(png);
  png_set_expand(png);
  png_set_strip_alpha(png);
  png_set_packing(png);
}

struct image * LoadPNG(const char *filename)
{
  struct image *img;
  FILE *fp;
  png_struct *png;
  png_info *png_info;
  long height;
  long row_size;
  int i;

  if (!(fp = fopen(filename, "r"))) {
    fprintf(stderr, "Error load \"%s\"\n", filename);
    return NULL;
  }

  if (!ispng(fp)) {
    fprintf(stderr, "Not png file: %s\n", filename);
    fclose(fp);
    return NULL;
  }

  if (!(png = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)) ||
      !(png_info = png_create_info_struct(png))) {
    fprintf(stderr, "png_create failed\n");
    fclose(fp);
    return NULL;
  }

  if (setjmp(png_jmpbuf(png))) {
    fprintf(stderr, "pnglib failed\n");
    png_destroy_read_struct(&png, &png_info, NULL);
    fclose(fp);
    return NULL;
  }

  png_init_io(png, fp);
  png_read_info(png, png_info);

  img = malloc(sizeof(struct image));

  img->res = png_get_pixels_per_meter(png, png_info);
  img->width = png_get_image_width(png, png_info);
  height = png_get_image_height(png, png_info);
  img->height = height - SKIP_TOP_ROWS - SKIP_BOTTOM_ROWS;
  img->xoff = 0;
  img->yoff = 0;

  png_set_filters(png);

  row_size = img->width * 3;

  img->data = malloc(img->height * sizeof(u8 *));
  for(i=0; i<img->height; i++)
    img->data[i] = malloc(row_size);

  png_read_rows(png, img->data, NULL, SKIP_TOP_ROWS);
  png_read_rows(png, img->data, NULL, img->height);

  png_destroy_read_struct(&png, &png_info, NULL);
  fclose(fp);

  deb("\"%s\" width=%ld height=%ld\n", filename, img->width, img->height);
  return img;

}

void DeletePNGrows (struct image *img) 
{
  long i;
  for (i = 0; i < img->height; i++)
    free(img->data[i]);
  free(img->data);
  img->data = NULL;
}

/* fall back when colors count > 256 */
static int palette_full(struct image *img, int x, int y, png_color *palette) {      
  deb("palette full, doing fall back%s\n", "");
  for(; y >= 0; y--)
  {   
    for(; x >= 0; x--)
    {
      u8 id = img->data[y][x];
      img->data[y][x*3] = palette[id].red;
      img->data[y][x*3+1] = palette[id].green;
      img->data[y][x*3+2] = palette[id].blue;
    }
    x = img->width-1;
  }
  return -1;
}

struct _palette {
  u32 key;
  int next; /* link to next color of same hash */
}; 

#define C(_r,_g,_b) ((_r<<16) | (_g<<8) | _b)

static int make_palette(struct image *img, png_color *palette)
{
  int c=0;
  long x, xx, y, i;
  struct _palette colors[256]; /* sync with *palette */
  int colors_hash[256];

  for(i=0; i<256; i++)
    colors_hash[i] = -1;

  for(y = 0; y < img->height; y++)
  {
    for(x = xx = 0; x < img->width; x++, xx+=3)
    {
      u8 red   = img->data[y][xx];
      u8 green = img->data[y][xx+1];
      u8 blue  = img->data[y][xx+2];
      u8 hash = red ^ green ^ blue;
      u32 key = C(red, green, blue);

      if ((i = colors_hash[hash]) < 0)
      {
        if (c==256)
          return palette_full(img, x-1, y, palette);

        colors_hash[hash] = c;
        palette[c].red = red;
        palette[c].green = green;
        palette[c].blue = blue;
        colors[c].key = key;
        colors[c].next = -1;
        deb("New color[%d] %d/%d/%d (hash %02x) at x=%ld y=%ld\n", c, red, green, blue, hash, x, y);
        i = c++;
      }
      else
      {
        while(colors[i].key != key)
        {
          if (colors[i].next < 0)
          {
            if (c==256)
              return palette_full(img, x-1, y, palette);

            colors[i].next = c;
            palette[c].red = red;
            palette[c].green = green;
            palette[c].blue = blue;
            colors[c].key = key;
            colors[c].next = -1;
            i = c++;
            break;
          }
          i = colors[i].next;
        }
      }
      img->data[y][x] = i;
    }
  }

  return c;
}

int SavePNG(struct image *img, const char *filename)
{
  FILE *fp;
  png_struct *png;
  png_info *png_info;
  png_color palette[256];
  int palette_count;

  deb("Save %s width=%ld height=%ld\n", filename, img->width, img->height);

  if (!(fp = fopen(filename, "wb"))) {
    fprintf(stderr, "Error open for write \"%s\"\n", filename);
    fclose(fp);
    return -1;
  }

  if (!(png = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)) ||
      !(png_info = png_create_info_struct(png))) {
    fprintf(stderr, "png_create failed\n");
    fclose(fp);
    return -1;
  }

  if (setjmp(png_jmpbuf(png))) {
    fprintf(stderr, "pnglib failed\n");
    png_destroy_write_struct(&png, &png_info);
    fclose(fp);
    return -1;
  }

  /* try palette */
  palette_count = make_palette(img, palette);

  png_init_io(png, fp);
  png_set_IHDR(png, png_info, 
      img->width, img->height,
      8, (palette_count > 0) ? PNG_COLOR_TYPE_PALETTE : PNG_COLOR_TYPE_RGB,
      PNG_INTERLACE_NONE,
      PNG_COMPRESSION_TYPE_DEFAULT,
      PNG_FILTER_TYPE_DEFAULT);

  if (palette_count > 0) {
    deb("Using palette count %d\n", palette_count);
    png_set_PLTE(png, png_info, palette, palette_count);
  }

  if (img->res > 0)
    png_set_pHYs(png, png_info, img->res, img->res, PNG_RESOLUTION_METER);

  png_set_rows(png, png_info, img->data);
  png_write_png(png, png_info, PNG_TRANSFORM_IDENTITY, NULL);

  png_destroy_write_struct(&png, &png_info);
  fclose(fp);

  return 0;

}

